# setuptools extension for building non-Python Dynamic Shared Objects

Building non-python shared libraries (eg. `libY.so`, `libY.dylib`, or `Y.dll`) for inclusion in a Python Wheel.

This extension provides at alternative to bundling externally built
libraries in Python Wheel packages.  This replaces an external
build system (eg. Makefile), allowing non-python libraries to be
built from source within the python ecosystem.

See [example/setup.py](example/setup.py) for a working example.

[![setuptools-dso](https://github.com/mdavidsaver/setuptools_dso/workflows/setuptools-dso/badge.svg)](https://github.com/mdavidsaver/setuptools_dso/actions)

## Building packages

Packages using `setuptools_dso` can for the most part be treated like any other python package.
Distribution through [pypi.org](https://pypi.org/) as source and/or binary (wheels) is possible.
Like other packages containing compiled code, use of egg binaries is not supported.

Use of PIP is encouraged.
The `setuptools_dso` package needs to be installed when manually invoking `setup.py` scripts in dependent packages.
eg. to generate source tars with `setup.py sdist`.

Developers wishing to work with an in-place build will need to use the added `build_dso` command,
which functions like the `build_ext` command.

```sh
python setup.py build_dso -i
python setup.py build_ext -i
```

## Usage

The [example/](example/) demonstrates building a non-python library,
and linking it with a python extension module.

### pyproject.toml

To properly support `pip install ...`, it is recommended to include a
[`pyproject.toml`](https://www.python.org/dev/peps/pep-0518/)
file containing at least:

```
[build-system]
requires = ["setuptools", "wheel", "setuptools_dso"]
```

This ensures that `setuptools_dso` is available to be imported by `setup.py`.

### MANIFEST.in

Add a `MANIFEST.in` to ensure that `setup.py sdist` includes everything necessary
for a successful source build.

```
include pyproject.toml
include src/*.h
include src/*.c
include src/*.cpp
```

### Building a DSO

The [source](example/src/) files while make up the non-python `demo` library are: `mylib.h`, `foo.c`, `bar.cpp`.
This library will be expressed as a `setuptools_dso.DSO` object.
The first argument is a directory path and library base name encoded like a python module.
eg. the result of `dsodemo.lib.demo` will be eg. `dsodemo/lib/libdemo.so` or `dsodemo\lib\demo.dll`
installed in the python module tree along-side any other python code or C extensions.

Note that there need not be a `dsodemo/lib/__init__.py` as `dsodemo.lib` need not be a python package.
However, if this file is present, then the generated `dsodemo/lib/demo_dsoinfo.py` will be accessible.

```py
from setuptools_dso import DSO, Extension, setup

dso = DSO('dsodemo.lib.demo', ['src/foo.c', 'src/bar.cpp'], ...)

setup(
    ...
    x_dsos = [dso],
    zip_safe = False, # setuptools_dso is not compatible with eggs!
)
```

The `DSO` constructor understands the same keyword arguments as `setuptools.Extension`
and [`distutils.core.Extension`](https://docs.python.org/3/distutils/apiref.html#distutils.core.Extension),
with the addition of `dsos=[...]`, `soversion='...'`, and `lang_compile_args={'...':'...'}`.

The `dsos=` argument is a list of other `DSO` names (eg. `'dsodemo.lib.demo'`) to allow
one `DSO` to be linked against others.

eg. `dsos=['some.lib.foo']` will result in something like `gcc ... -L.../some/lib -lfoo`.

### Building an Extension

`setuptools_dso.Extension` is a wrapper around `setuptools.Extension` which adds the `dsos=[...]` keyword argument.
This allows a C extension module to be linked against a `DSO` by name.
The named `DSO` may be built by the same `setup.py`, or may already be present in `$PYTHONPATH`.

```py
from setuptools_dso import DSO, Extension, setup

ext = Extension('dsodemo.ext.dtest', ['src/extension.cpp'],
    dsos=['dsodemo.lib.demo'],
)

setup(
    ...
    ext_modules = [ext],
    zip_safe = False, # setuptools_dso is not compatible with eggs!
)
```

### Cython

Version 1.3 added a `setuptools_dso.cythonize()` wrapper to correctly handle `Extension(dso=...)`.

### Runtime

Beginning with setuptools-dso 2.0 a file `*_dsoinfo.py` will be generated alongside each DSO.
eg. dso `mypkg.lib.thelib` will create `mypkg/lib/thelib_dsoinfo.py`.
If `mypkg.lib` is a valid python packages (contains `__init__.py`)
then `mypkg.lib.thelib_dsoinfo` can be imported, and will contain the attributes:

- `.dsoname`    eg. "mypkg.lib.thelib"
- `.libname`    eg. "thelib.so"
- `.soname`     eg. "thelib.so.0"
- `.filename`   eg. "/full/path/to/thelib.so"
- `.sofilename` eg. "/full/path/to/thelib.so.0"

Beginning with 2.0 `os.add_dll_directory()` is called by the generated `*_dsoinfo.py` module.
Prior to 2.0, or if the generated module is not used,
some additional runtime preparation is needed in order to find the `'dsodemo.lib.demo'` library
when the `dsodemo.ext.dtest` Extension is imported on Windows.
This could be placed in eg. `example/src/dsodemo/__init__.py`
to ensure it always runs before the extension library is loaded.

```py
import sys, os

def fixpath():
    # If this file is
    #   .../__init__.py
    # DSOs are under
    #   .../lib/
    libdir = os.path.join(os.path.dirname(__file__), 'lib')

    if hasattr(os, 'add_dll_directory'):
        os.add_dll_directory(libdir)
    else:
        path = os.environ.get('PATH', '').split(os.pathsep)
        path.append(libdir)
        os.environ['PATH'] = os.pathsep.join(path)

if sys.platform == "win32":
    fixpath()
```

## Mechanics

Libraries build with setuptools-dso will typically be linked, directly or indirectly, to python extension modules.
Such libraries should be loaded implicitly when the extension module is imported by the python runtime.
For this to work the OS runtime loader must be able to find these libraries.
This is complicated by OS differences, and a desire to support virtualenv and similar.

Supporting virtualenv prevents the use of absolute build time paths.

The resolutions to this complication differ for each OS depending on the
capabilities of the executable format used.

### ELF (Linux, *NIX, *BSD)

For ELF (Executable and Linking Format) targets the RPATH mechanism is used with the magic $ORIGIN/ prefix.

Take as an example the following installed module tree:

```
* modulea
 * __init__.py
 * _ext.so
 * lib
  * libsup.so
  * libsup.so.0
```

The supporting library is then linked with a library name of `libsup.so.0`
(eg. `gcc -o libsup.so.0 -shared -Wl,-soname,libsup.so.0 ...`)

When linking a dependent library or extension module `-Wl,-rpath,'$ORIGIN/lib'`
is given since the relative path from `_ext.so` to `libsup.so.0` is `./lib`.

Note that on Linux runtime linking is really a function of the libc (eg. glibc).

### Mach-O (OSX)

Historically Mach-O did not support anything like RPATH.
However, it now does (cf. 'man dyld')

The equivalence with GCC+ELF is for `-Wl,-soname,libsup.so.0` to become `-install\_name @rpath/libsup.0.dylib`.
Then `-Wl,-rpath,'$ORIGIN/lib'` becomes `-Wl,-rpath,@loader_path/lib`.

### PE (Windows)

There is no equivalent for RPATH.
It is necessary to adjust the loader search path to include the directory
containing the support library dll explicitly from python code prior to loading the extension.
(eg. from an `__init__.py`)

With earlier versions of Windows it is sufficient to adjust `%PATH%`.
Recent versions require use of the `AddDllDirectory()` API call.
Python wraps this as `os.add_dll_directory()`.

## Changelog

### 2.0 (UNRELEASED)

- Generate info module for each DSO (Kaleb Barrett)

### 1.7 (May 2020)

- Fix support for develop install (pip install -e).  (Kirill Smelkov)

### 1.6 (May 2020)

- Use @rpath as install_name prefix on MacOS (Mehul Tikekar)

### 1.5 (Jan. 2020)

- Correct parallel compile on Mac

### 1.4 (Jan. 2020)

- Take package_dir into account (Kirill Smelkov)
- Parallel compile of DSO source when supported (py >= 3.4 and not windows)
- Fake `cythonize()` wrapper when Cython not installed

### 1.3 (Nov. 2019)

- Handle DSO in root (not in a package)
- setup() wrapper correct 'cmdclass' handling (Kirill Smelkov)
- Add `setuptools_dso.cythonize()` wrapper

### 1.2 (Aug. 2019)

- Add LICENSE and COPYRIGHT to MANIFEST

### 1.1 (Aug. 2019)

- Fix inplace build (Kirill Smelkov)

### 1.0 (Oct. 2018)

- Initial Release
